package riven.virtualthreads;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;
import java.util.PriorityQueue;
import java.util.Queue;

/*
 * Created on Nov 30, 2010
 * 
 * @author Riven
 */

public class CopyOfVirtualProcessor {
   
   static final ThreadLocal<CopyOfVirtualProcessor> LAST_DEFINED = new ThreadLocal<CopyOfVirtualProcessor>();
   
   private final Queue<VirtualThread>               queue;
   private final List<VirtualThread>                later;
   
   public CopyOfVirtualProcessor() {
      this.queue = new PriorityQueue<VirtualThread>(11, new VirtualThreadComparator());
      this.later = new ArrayList<VirtualThread>();
      
      LAST_DEFINED.set(this);
   }
   
   private long currentTickTimestamp;
   
   public final long getCurrentTime() {
      return this.currentTickTimestamp;
   }
   
   private VirtualThread current;
   
   /**
    * Executes the virtual threads that are scheduled, if any.
    * The new threads, or threads that are resumed or reschedule
    * themselves to be executes will be executed in the next tick.  
    */
   
   public boolean tick(long now) {
      
      this.currentTickTimestamp = now;
      
      this.sync();
      
      while (true) {
         VirtualThread thread;
         
         thread = this.queue.peek();
         if (thread == null || thread.wakeUpAt > now) {
            break;
         }
         thread = this.queue.poll();
         
         this.current = thread;
         VirtualThreadState state = thread.step();
         this.current = null;
         
         if (state == VirtualThreadState.RUNNABLE) {
            this.later.add(thread);
         }
      }
      
      return !this.queue.isEmpty() || !this.later.isEmpty();
   }
   
   boolean unschedule(VirtualThread thread) {
      if (this.current == thread) {
         return true;
      }
      if (this.later.remove(thread)) {
         return true;
      }
      return this.queue.remove(thread);
   }
   
   void enqueue(VirtualThread thread) {
      this.later.add(thread);
   }
   
   void sync() {
      if (!this.later.isEmpty()) {
         this.queue.addAll(this.later);
         this.later.clear();
      }
   }
   
   static class VirtualThreadComparator implements Comparator<VirtualThread> {
      @Override
      public int compare(VirtualThread o1, VirtualThread o2) {
         int val = Long.signum(o1.wakeUpAt - o2.wakeUpAt);
         return (val == 0) ? +1 : val;
      }
   }
}
