/*
 * Created on May 27, 2012
 */

package riven.virtualthreads.test;

import java.awt.Color;
import java.awt.Dimension;
import java.util.ArrayList;
import java.util.List;

import javax.swing.JFrame;
import javax.swing.SwingUtilities;

import riven.virtualthreads.VirtualProcessor;

public class Game {
   public static long                 TIME          = now();
   public static final float          TICK_DURATION = 0.01f;
   
   public static final List<GameItem> ITEMS0        = new ArrayList<GameItem>();
   
   public static VirtualProcessor     PROCESSOR;
   
   public static long now() {
      return System.nanoTime() / 1000000L;
   }
   
   public static void addItem(GameItem item) {
      ITEMS0.add(item);
      item.onAdd();
   }
   
   public static void removeItem(GameItem item) {
      ITEMS0.remove(item);
      item.onRemove();
   }
   
   public static void main(String[] args) throws Exception {
      
      SwingUtilities.invokeAndWait(new Runnable() {
         @Override
         public void run() {
            init();
         }
      });
      
      final GameCanvas canvas = new GameCanvas();
      
      SwingUtilities.invokeAndWait(new Runnable() {
         @Override
         public void run() {
            
            canvas.setPreferredSize(new Dimension(1024, 768));
            
            JFrame holder = new JFrame();
            holder.getContentPane().add(canvas);
            holder.setResizable(false);
            holder.pack();
            holder.setLocationRelativeTo(null);
            holder.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
            holder.setVisible(true);
         }
      });
      
      while (true) {
         SwingUtilities.invokeAndWait(new Runnable() {
            @Override
            public void run() {
               tick();
            }
         });
         
         canvas.repaint();
         
         Thread.sleep(10);
      }
   }
   
   static StaticGameItemGrid TREE_GRID    = new StaticGameItemGrid(1024, 768, 24);
   static StaticGameItemGrid WELL_GRID    = new StaticGameItemGrid(1024, 768, 128);
   static StaticGameItemGrid STORAGE_GRID = new StaticGameItemGrid(1024, 768, 128);
   
   public static void init() {
      PROCESSOR = new VirtualProcessor();
      
      if (false) {
         GameItem player = new GameItem();
         player.name = "Player";
         player.color = Color.BLUE;
         player.position.set(256.0f, 256.0f);
         player.speed = 60.0f;
         Game.addItem(player);
      }
      
      // add trees
      for (int i = 0; i < 1024 / 4; i++) {
         GameItemTree tree = new GameItemTree();
         tree.position.x = Misc.random(32, 1024 - 32);
         tree.position.y = Misc.random(32, 768 - 32);
         Game.addItem(tree);
      }
      
      // add storages
      for (int i = 0; i < 7; i++) {
         GameItemStorage storage = new GameItemStorage();
         storage.position.x = Misc.random(32, 1024 - 32);
         storage.position.y = Misc.random(32, 768 - 32);
         Game.addItem(storage);
      }
      
      // add lumberjacks
      for (int i = 0; i < 25 * 0; i++) {
         GameItemLumberjack lumberjack = new GameItemLumberjack();
         lumberjack.position.x = Misc.random(32, 1024 - 32);
         lumberjack.position.y = Misc.random(32, 768 - 32);
         Game.addItem(lumberjack);
      }
      
      // add water
      for (int i = 0; i < 5; i++) {
         GameItemWell water = new GameItemWell();
         water.position.x = Misc.random(32, 1024 - 32);
         water.position.y = Misc.random(32, 768 - 32);
         Game.addItem(water);
      }
   }
   
   public static void tick() {
      Game.TIME = now();
      
      long t0 = now();
      int tasks = Game.PROCESSOR.tick(Game.TIME);
      long t1 = now();
      
      //  System.out.println(Game.ITEMS0.size() + " items, " + tasks + " tasks, took: " + (t1 - t0));
   }
}
