/*
 * Created on May 27, 2012
 */

package riven.virtualthreads.test;

import java.awt.Color;

import craterstudio.func.Filter;

import riven.virtualthreads.VirtualRunnable;
import riven.virtualthreads.VirtualThread;

import de.matthiasmann.continuations.SuspendExecution;

public class GameItemLumberjack extends GameItem {
   
   public GameItemLumberjack() {
      this.name = null;//"lumberjack";
      this.color = Color.DARK_GRAY;
      
      this.speed = 25.0f + (float) Math.random() * 25.0f;
      
      resources.register(ResourceType.WOOD, new ResourceHolder(Misc.random(5, 8)));
      resources.register(ResourceType.WATER, new ResourceHolder(50, Misc.random(50, 100)));
      
      this.spawn(new VirtualRunnable() {
         @Override
         public void run() throws SuspendExecution {
            GameItemLumberjack.this.ai();
         }
      });
      
      this.spawn(new VirtualRunnable() {
         @Override
         public void run() throws SuspendExecution {
            while (true) {
               VirtualThread.sleep(1000);
               resources.get(ResourceType.WATER).take(1);
            }
         }
      });
   }
   
   public int chop(GameItemTree tree) throws SuspendExecution {
      
      int transfer = ResourceGroup.transfer(ResourceType.WOOD, tree.resources, this.resources, 1);
      
      if (tree.resources.get(ResourceType.WOOD).isEmpty()) {
         tree.kill();
      }
      
      return transfer;
   }
   
   private void humanAI() throws SuspendExecution {
      Vec2 goBackTo = new Vec2(this.position);
      while (this.resources.get(ResourceType.WATER).getRatio() < 0.33f) {
         GameItemWell water = this.findNearestWater();
         if (water == null) {
            break;
         }
         
         moveTo(water.position);
         
         ResourceGroup.transfer(ResourceType.WATER, water.resources, this.resources);
         
         VirtualThread.sleep(Misc.random(500, 1500));
         
         if (this.resources.get(ResourceType.WATER).isEmpty()) {
            System.out.println("lumberjack died of thirst");
            this.kill();
            throw new IllegalStateException("never happens");
         }
      }
      moveTo(goBackTo);
   }
   
   private GameItemTree gatherWood() throws SuspendExecution {
      GameItemTree lastTree = null;
      while (!this.resources.get(ResourceType.WOOD).isFull()) {
         
         long t0 = System.nanoTime();
         GameItemTree tree = this.findNearestTree();
         long t1 = System.nanoTime();
         System.out.println("searching tree took: " + (t1 - t0));
         
         if (tree == null) {
            break;
         }
         this.moveTo(tree.position);
         this.humanAI();
         
         while (this.chop(tree) > 0) {
            VirtualThread.sleep(Misc.random(500, 1500));
            this.humanAI();
         }
         
         lastTree = tree;
      }
      return lastTree;
   }
   
   private void dumpWood() throws SuspendExecution {
      while (!this.resources.get(ResourceType.WOOD).isEmpty()) {
         GameItemStorage storage = this.findNearestStorageWithFreeSpace();
         if (storage == null) {
            break;
         }
         this.moveTo(storage.position);
         this.humanAI();
         
         int dumped = ResourceGroup.transfer(ResourceType.WOOD, this.resources, storage.resources);
         
         VirtualThread.sleep(1000);
      }
   }
   
   public void ai() throws SuspendExecution {
      
      GameItemTree lastTree = null;
      
      // forever
      while (true) {
         Vec2 goBackTo = new Vec2(this.position);
         while (this.resources.get(ResourceType.WATER).getRatio() < 0.33f) {
            GameItemWell water = this.findNearestWater();
            if (water == null) {
               break;
            }
            
            moveTo(water.position);
            
            ResourceGroup.transfer(ResourceType.WATER, water.resources, this.resources);
            
            VirtualThread.sleep(Misc.random(500, 1500));
            
            if (this.resources.get(ResourceType.WATER).isEmpty()) {
               System.out.println("lumberjack died of thirst");
               this.kill();
               throw new IllegalStateException("never happens");
            }
         }
         moveTo(goBackTo);
         
         //
         
         while (!this.resources.get(ResourceType.WOOD).isFull()) {
            
            long t0 = System.nanoTime();
            GameItemTree tree = this.findNearestTree();
            long t1 = System.nanoTime();
            System.out.println("searching tree took: " + (t1 - t0));
            
            if (tree == null) {
               break;
            }
            this.moveTo(tree.position);
            this.humanAI();
            
            while (this.chop(tree) > 0) {
               VirtualThread.sleep(Misc.random(500, 1500));
               this.humanAI();
            }
            
            lastTree = tree;
         }
         
         //
         
         while (!this.resources.get(ResourceType.WOOD).isEmpty()) {
            GameItemStorage storage = this.findNearestStorageWithFreeSpace();
            if (storage == null) {
               break;
            }
            this.moveTo(storage.position);
            this.humanAI();
            
            int dumped = ResourceGroup.transfer(ResourceType.WOOD, this.resources, storage.resources);
            
            VirtualThread.sleep(1000);
         }
         
         //
         
         if (lastTree != null) {
            this.moveTo(lastTree.position);
         }
         
         VirtualThread.sleep(1000);
      }
   }
   
   public void aiNew() throws SuspendExecution {
      
      while (true) {
         this.humanAI();
         
         this.gatherWood();
         this.dumpWood();
         
         VirtualThread.sleep(1000);
      }
   }
   
   private void moveTo(Vec2 pos) throws SuspendExecution {
      MoveAI.moveTo(this, pos);
   }
   
   private GameItemWell findNearestWater() {
      return (GameItemWell) Game.WELL_GRID.scan(this.position, 1000.0f, new Filter<GameItem>() {
         @Override
         public boolean accept(GameItem item) {
            return !((GameItemWell) item).resources.get(ResourceType.WATER).isEmpty();
         }
      });
   }
   
   private GameItemTree findNearestTree() throws SuspendExecution {
      for (float range = 100.0f; range < 1000.0f; range *= 1.33f) {
         GameItemTree tree = (GameItemTree) Game.TREE_GRID.scan(this.position, range, null);
         if (tree != null) {
            return tree;
         }
         VirtualThread.sleep(2 * 1000);
      }
      return null;
   }
   
   private GameItemStorage findNearestStorageWithFreeSpace() {
      return (GameItemStorage) Game.STORAGE_GRID.scan(this.position, 1000.0f, new Filter<GameItem>() {
         @Override
         public boolean accept(GameItem item) {
            return !((GameItemStorage) item).resources.get(ResourceType.WOOD).isFull();
         }
      });
   }
}
