/*
 * Created on May 27, 2012
 */

package riven.virtualthreads.test;

public class ResourceHolder {
   private int holding;
   private int capacity;
   
   public ResourceHolder(int capacity) {
      if (capacity < 1) {
         throw new IllegalArgumentException();
      }
      this.capacity = capacity;
   }
   
   public ResourceHolder(int holding, int capacity) {
      if (capacity < 1) {
         throw new IllegalArgumentException();
      }
      if (holding < 0) {
         throw new IllegalArgumentException();
      }
      if (holding > capacity) {
         throw new IllegalArgumentException();
      }
      this.holding = holding;
      this.capacity = capacity;
   }
   
   public void grow(int amount) {
      if (amount < 0) {
         throw new IllegalArgumentException();
      }
      this.capacity += amount;
   }
   
   public void shrink(int amount) {
      if (amount < 0) {
         throw new IllegalArgumentException();
      }
      if (this.capacity - amount < this.holding) {
         throw new IllegalStateException();
      }
      this.capacity -= amount;
   }
   
   public boolean isEmpty() {
      return this.holding == 0;
   }
   
   public float getRatio() {
      return (float) this.holding / this.capacity;
   }
   
   public boolean isFull() {
      return this.holding == this.capacity;
   }
   
   public int getFilledSpace() {
      return this.holding;
   }
   
   public int getFreeSpace() {
      return this.capacity - this.holding;
   }
   
   public int put(int amount) {
      if (amount < 0) {
         throw new IllegalArgumentException();
      }
      int transfer = Math.min(amount, this.getFreeSpace());
      this.holding += transfer;
      return transfer;
   }
   
   public int take(int amount) {
      if (amount < 0) {
         throw new IllegalArgumentException();
      }
      int transfer = Math.min(amount, this.holding);
      this.holding -= transfer;
      return transfer;
   }
   
   @Override
   public String toString() {
      return this.holding + "/" + this.capacity;
   }
   
   //
   
   public static int transfer(ResourceHolder src, ResourceHolder dst) {
      return transfer(src, dst, Integer.MAX_VALUE);
   }
   
   public static int transfer(ResourceHolder src, ResourceHolder dst, int max) {
      int transfer = Math.min(src.getFilledSpace(), dst.getFreeSpace());
      
      transfer = Math.min(max, transfer);
      
      if (src.take(transfer) != transfer) {
         throw new IllegalStateException();
      }
      
      if (dst.put(transfer) != transfer) {
         throw new IllegalStateException();
      }
      
      return transfer;
   }   
}