package riven.virtualthreads.util;

import java.lang.instrument.ClassFileTransformer;
import java.lang.instrument.IllegalClassFormatException;
import java.lang.instrument.Instrumentation;
import java.security.ProtectionDomain;

import org.objectweb.asm.ClassReader;
import org.objectweb.asm.ClassVisitor;
import org.objectweb.asm.ClassWriter;
import org.objectweb.asm.util.CheckClassAdapter;

import de.matthiasmann.continuations.instrument.*;

/*
 * Created on Nov 21, 2010
 * 
 * @author Riven
 */

public class VirtualThreadAgent {
   public static void premain(String agentArguments, Instrumentation instrumentation) {
      final MethodDatabase db = new MethodDatabase(Thread.currentThread().getContextClassLoader());
      
      db.setVerbose(false);
      db.setDebug(false);
      db.setAllowMonitors(false);
      
      boolean checkArg = false;
      
      if (agentArguments != null) {
         for (char c : agentArguments.toCharArray()) {
            switch (c) {
               case 'v':
                  db.setVerbose(true);
                  break;
               
               case 'd':
                  db.setDebug(true);
                  break;
               
               case 'm':
                  db.setAllowMonitors(true);
                  break;
               
               case 'c':
                  checkArg = true;
                  break;
               
               default:
                  throw new IllegalStateException("Usage: vdmc (verbose, debug, allow monitors, check class)");
            }
         }
      }
      
      final boolean check = checkArg;
      
      db.setLog(new Log() {
         public void log(String msg, Object... args) {
            if (db.isVerbose()) {
               System.out.println("[GreenThreads] LOG: " + String.format(msg, args));
            }
         }
         
         public void error(String msg, Exception exc) {
            if (db.isVerbose()) {
               System.out.println("[GreenThreads] ERROR: " + msg);
               
               exc.printStackTrace(System.out);
            }
         }
      });
      
      instrumentation.addTransformer(new ClassFileTransformer() {
         @Override
         public byte[] transform(ClassLoader loader, String className, Class< ? > classBeingRedefined, ProtectionDomain protectionDomain, byte[] classfileBuffer) throws IllegalClassFormatException {
            if (className.startsWith("java/"))
               return null;
            if (className.startsWith("javax/"))
               return null;
            if (className.startsWith("sun/"))
               return null;
            
            long t0 = System.nanoTime();
            byte[] bytecode = instrumentClass(db, classfileBuffer, check);
            long t1 = System.nanoTime();
            
            if (db.isVerbose()) {
               System.out.println("[GreenThreads] TRANSFORM: " + className + " (" + ((t1 - t0) / 1000000L) + "ms)");
            }
            
            return bytecode;
         }
      });
   }
   
   static byte[] instrumentClass(MethodDatabase db, byte[] data, boolean check) {
      ClassReader r = new ClassReader(data);
      ClassWriter cw = new ClassWriter(ClassWriter.COMPUTE_FRAMES | ClassWriter.COMPUTE_MAXS);
      ClassVisitor cv = check ? new CheckClassAdapter(cw) : cw;
      InstrumentClass ic = new InstrumentClass(cv, db);
      r.accept(ic, ClassReader.SKIP_FRAMES);
      return cw.toByteArray();
   }
}