/*
 * Created on May 27, 2012
 */

package riven.virtualthreads.util;

import java.util.concurrent.atomic.AtomicInteger;

import riven.virtualthreads.VirtualRunnable;
import riven.virtualthreads.VirtualThread;
import de.matthiasmann.continuations.SuspendExecution;

public class VirtualTrigger implements Runnable {
   
   private final VirtualThread thread;
   
   public VirtualTrigger() {
      this.thread = VirtualThread.currentThread();
   }
   
   @Override
   public void run() {
      this.fire();
   }
   
   private static final int PENDING = 0;
   private static final int FIRED   = 1;
   private static final int TIMEOUT = 2;
   
   private AtomicInteger    state   = new AtomicInteger(PENDING);
   
   boolean hasFired() {
      return this.state.get() != PENDING;
   }
   
   public void fire() {
      if (this.state.compareAndSet(PENDING, FIRED)) {
         this.thread.resume();
      }
   }
   
   public void timeout() {
      if (this.state.compareAndSet(PENDING, TIMEOUT)) {
         this.thread.resume();
      }
   }
   
   //
   
   public long await() throws SuspendExecution {
      return await(0);
   }
   
   public long await(final long timeout) throws SuspendExecution {
      
      if (timeout < 0) {
         throw new IllegalArgumentException();
      }
      
      if (this.hasFired()) {
         return 0;
      }
      
      if (timeout > 0) {
         this.thread.spawn(new VirtualRunnable() {
            @Override
            public void run() throws SuspendExecution {
               VirtualThread.sleep(timeout);
               VirtualTrigger.this.timeout();
            }
         });
      }
      
      long started = this.thread.getProcessor().getCurrentTime();
      
      VirtualThread.suspend();
      
      long elapsed = this.thread.getProcessor().getCurrentTime() - started;
      
      switch (this.state.get()) {
         case PENDING:
            throw new IllegalStateException();
            
         case FIRED:
            return elapsed;
            
         case TIMEOUT:
            if (elapsed == 0) {
               // ensure negative value
               return -1;
            }
            return -elapsed;
            
         default:
            throw new IllegalStateException("never happens");
      }
   }
}
